#!/usr/bin/env python3
from __future__ import print_function
import contextlib
import base64
import datetime
from functools import wraps
import glob
import itertools
import logging
from logging.handlers import WatchedFileHandler
from operator import attrgetter, itemgetter
import os
import re
import shutil
import subprocess
import sys
import tempfile
import time
import uuid
import zipfile
import zlib
import random
try:
    import geoip2
    import geoip2.database
except ImportError:
    geoip2 = None

from dateutil.parser import parse as parse_dt
import psutil
import psycopg2
from psycopg2.extensions import QuotedString
from psycopg2.extras import NamedTupleCursor
from psycopg2.pool import ThreadedConnectionPool
import requests
try:
    import urllib3
except ImportError:
    from requests.packages import urllib3
if hasattr(urllib3, 'disable_warnings'):  # noqa
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)  # noqa
import json
from dateutil import tz

try:
    from setproctitle import setproctitle
except ImportError:
    setproctitle = lambda t: None

try:
    from auth_util import SaasAuth
    from const import (
        LOCAL,
        TEST,
        USER,
        PG_ODOO_PASS,
        PG_USER,
        PG_PASS,
        METABASE_PG_HOST,
        SCHEME,
        SAAS_BASE_URL,
        MASTER,
        BRANCH_ALIASES,
        EDU_PREFIX,
        SAAS_VALID_SUFFIXES,
        TIMEOUT,
        TMPMIGDIR,
        TMPMIGSTATICDIR,
        TMPRESTOREDIR,
        WEBSITE,
    )
    from upgrade_api import get_upgrade_api, generate_token, UpgradeError
    from upgrade_api import _logger as _logger_api
except ImportError:
    from .auth_util import SaasAuth
    from .const import (
        LOCAL,
        TEST,
        USER,
        PG_ODOO_PASS,
        PG_USER,
        PG_PASS,
        METABASE_PG_HOST,
        SCHEME,
        SAAS_BASE_URL,
        MASTER,
        BRANCH_ALIASES,
        EDU_PREFIX,
        SAAS_VALID_SUFFIXES,
        TIMEOUT,
        TMPMIGDIR,
        TMPMIGSTATICDIR,
        TMPRESTOREDIR,
        WEBSITE,
    )
    from .upgrade_api import get_upgrade_api, generate_token, UpgradeError
    from .upgrade_api import _logger as _logger_api

MAX_CONN = 1024
ODOO_SRC = os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', '..', '..', 'odoo'))
try:
    import pwd
    # use effective uid home dir for munin probe
    LOG_DIR = os.path.join(pwd.getpwuid(os.geteuid()).pw_dir, 'log')
    try:
        os.mkdir(LOG_DIR)
    except OSError:
        pass
except ImportError:
    LOG_DIR = '.'       # for designers on windows...

PID_DIR = LOG_DIR

COLUMNS = []
DT_COLUMNS = []

MAX_SELF_MIGRATION = 1

DB_DOMAIN_FIELDS = {
    'name',
    'https_redir',
}
DB_LIMIT_FIELDS = {
    'email_daily_limit',
    'email_daily_count',
    'ext_email_daily_count',
    'email_total_count',
    'print_waiting_counter',
    'print_counter',
    'print_counter_limit',
    'expire_date',
    'secure',
}
DB_STAT_FIELDS = {
    'date',
    'email_limit',
    'email_count',
    'ext_email_count',
    'bounce_count',
    'installed_app_count',
    'regular_user_count',
    'user_count',
}
MASTER_TIMEOUT = 120
RR_SCHEDULE = {
    '14.0': "November 11, 2023",
    '15.0': "October 11, 2024",
    'saas~15.2': "March 16, 2024",
    'saas~16.2': "March 16, 2024",
    'saas~16.3': "April 7, 2024",
    'saas~16.4': "May 12, 2024",
    'saas~17.1': "June 29, 2024",
}

_pool = None
_pg_server_version = None

pg_env = dict(os.environ, PGPASSWORD=PG_PASS) if PG_PASS else dict(os.environ)
pg_cnx = ("-h localhost -U " + PG_USER) if PG_PASS else ''

"""
   Modes:
       - trial: free trial
            - expire_date set (4h) before email confirmation
            - expire_date set (15d) after confirmation (+ 3mo on connection) (free for one app)
            - expire_date set (15d) after another app installation, extra_app set
       - paid: paid once, contract exists
       - drop: deactivated, expire_date set
       - spare: spare empty database
       - hostname: template line for the current host (usually unique)

   Statuses copied from contract when contract exists:
       - open: contract valid (mode=paid)
       - pending: automatic renewal failed, waiting for manual payment, mode=paid
       - close: contract closed (usually if stays pending during 15 days), mode=drop, expire_date set (3 months), nginx=402
       - cancel: contract cancelled manually, mode=drop, expire_date set (3 months), nginx=402
   Extra statuses:
       - frozen: trial database that has not been activated within the initial 4-hour window
       - unsubscribed: contract deleted (manually requested), mode=drop, expire_date set (3 months), nginx=402
       - maintenance*: temporarily down for migration or move to other host, mode=*, nginx=503
"""
ACTIVE_MODES = ('trial', 'paid')
DISABLED_STATUSES = ('close', 'cancel', 'unsubscribed', 'abuse', 'frozen',
                     'maintenance', 'maintenance_mig', 'maintenance_block')

"""
    Default email daily limits
"""
EMAIL_DAILY_LIMIT_PAID = 200
DUPLICATE_LIMIT_PAID = 10


# databases that should not ping after upgrade
_NO_PING = [
    "8851207e-1ff9-11e0-a147-001cc0f2115e",  # openerp
]

_logger = logging.getLogger(__name__)
_log_fmt = logging.Formatter('[%(asctime)s] %(levelname)s pid=%(process)d db=meta %(message)s')
_log_file = os.path.join(LOG_DIR, 'meta.log')
_log_hndl = WatchedFileHandler(_log_file)
_log_hndl.setFormatter(_log_fmt)
_logger.addHandler(_log_hndl)
_logger.setLevel(logging.INFO)
_logger_api.addHandler(_log_hndl)
_logger_api.setLevel(logging.INFO)

def log_call(f):
    @wraps(f)
    def w(*a, **kw):
        _logger.info('call %s(*%r, **%r)', f.__name__, a, kw)
        return f(*a, **kw)
    return w


# ----------------------------------------------------------------------------
# signal handling

_STOPPED = False
_STOPPABLE = False


def stopped():
    return _STOPPED


def graceful_stop(f):
    @wraps(f)
    def w(*a, **kw):
        global _STOPPABLE
        _STOPPABLE = True
        return f(*a, **kw)
    return w


def check_stop(it):
    for i in it:
        if _STOPPED:
            break
        yield i


def _sigint_handler(s, f):
    global _STOPPED
    if _STOPPABLE and not _STOPPED:
        _logger.info('CTRL-C catched. Will Stop after next operation')
        sys.stdout.flush()
        _STOPPED = True
    else:
        if _STOPPABLE:
            _logger.info('CTRL-C catched. Stop immediatly')
            sys.stdout.flush()
        raise KeyboardInterrupt()


# ----------------------------------------------------------------------------
# sql mgmt
@contextlib.contextmanager
def cursor(commit=True):
    global _pool, _pg_server_version, COLUMNS, DT_COLUMNS
    try:
        cnx = _pool.getconn()
    except AttributeError:
        dsn = 'dbname=meta'
        if not LOCAL:
            dsn += ' user=odoo password={} host={} sslmode=disable'.format(
                PG_ODOO_PASS,
                METABASE_PG_HOST
            )
        _pool = ThreadedConnectionPool(1, MAX_CONN, dsn)
        cnx = _pool.getconn()

        if not _pg_server_version:
            _pg_server_version = cnx.server_version

    cr = cnx.cursor(cursor_factory=NamedTupleCursor)

    # when run inside openerp, force the datetime fields to cast to python datetime
    def str2dt(value, cr):
        if value is None or isinstance(value, datetime.datetime):
            return value
        return parse_dt(value)
    for pg_code, type_name in ((1114, 'datetime'), (1082, 'date')):
        pgtyp = psycopg2.extensions.new_type((pg_code,), type_name, str2dt)
        psycopg2.extensions.register_type(pgtyp, cr)    # only for this cursor

    if not COLUMNS:
        # initialize column names
        cr.execute('SELECT * FROM databases,migrations WHERE 1=0')
        COLUMNS = list(map(attrgetter('name'), cr.description))
        DT_COLUMNS = set([c.name for c in cr.description if c.type_code in psycopg2.DATETIME.values])
    try:
        to_commit = False
        yield cr
        to_commit = commit
    finally:
        if to_commit:
            cnx.commit()
        else:
            try:
                cnx.rollback()
            except psycopg2.InterfaceError:
                # broken cnx, do not put it back in pool
                return
        try:
            cr.close()
        finally:
            _pool.putconn(cnx)

@contextlib.contextmanager
def cursor_or_new(cr=None):
    if cr:
        yield cr
    else:
        with cursor() as cr:
            yield cr

def isnamedtuple(o):
    return isinstance(o, tuple) and hasattr(o, "_replace") and hasattr(o, "_asdict")

def has_role(role):
    try:
        with open(os.path.expanduser('~/config/config.json')) as c:
            saas_config = json.load(c)
    except (IOError, ValueError):
        saas_config = {}

    return role in saas_config.get('me', {}).get('mode', '').split()

class sql(object):
    """Simple query builder"""
    def __init__(self, query, *values):
        self._query = query
        self._values = tuple(v for v in values if v is not None)    # only None. filter(None, values) drop false values

    def bind(self, column):
        return self._query.format(column), self._values

    @staticmethod
    def quote_dt(value):
        """quote datetime if not a datetime method"""
        if not value:
            return 'NULL'

        forbidden_keywords = '-- ; , AND OR UNION'.upper().split()
        u = value.upper()
        if u == 'NULL' or (all(f not in u for f in forbidden_keywords) and value.startswith("now()")):
            return value
        result = QuotedString(value).getquoted()
        # "In Python 3 the SQL must be returned in a bytes object. The ISQLQuote implementation does nothing."
        # but we're injecting these values straight into str contexts -> need to convert it back to str
        if bytes != str: # false in P2, true in P3 \o/
            return result.decode('utf-8')
        return result

    @classmethod
    def eq(cls, value, wildcard=False, caseinsensitive=False):
        if value is None:
            q = '{0} IS NULL'
        elif wildcard:
            op = 'ilike' if caseinsensitive else 'like'
            q = '{{0}} {0} %s'.format(op)
            value = '%{0}%'.format(value)
        elif caseinsensitive:
            q = 'lower({0}) = lower(%s)'
        else:
            q = '{0} = %s'

        return cls(q, value)

    @classmethod
    def neq(cls, value, wildcard=False, caseinsensitive=False):
        if value is None:
            q = '{0} IS NOT NULL'
        elif wildcard:
            op = 'not ilike' if caseinsensitive else 'not like'
            q = '({{0}} {0} %s OR {{0}} IS NULL)'.format(op)
            value = '%{0}%'.format(value)
        elif caseinsensitive:
            q = '(lower({0}) != lower(%s) or {0} IS NULL)'
        else:
            q = '({0} != %s OR {0} IS NULL)'

        return cls(q, value)

    @classmethod
    def lt(cls, value):
        return cls('{0} < %s', value)

    @classmethod
    def lte(cls, value):
        return cls('{0} <= %s', value)

    @classmethod
    def gt(cls, value):
        return cls('{0} > %s', value)

    @classmethod
    def gte(cls, value):
        return cls('{0} >= %s', value)

    @classmethod
    def between(cls, lower_bound, upper_bound):
        return cls('{0} BETWEEN %s AND %s', lower_bound, upper_bound)

    @classmethod
    def in_(cls, *value):
        match_null = ''
        if None in value:
            value = tuple(v for v in value if v is not None)
            match_null = ' OR {0} IS NULL'
        return cls('({0} IN %s' + match_null + ')', value)

    @classmethod
    def not_in(cls, *value):
        match_null = ''
        if None in value:
            value = tuple(v for v in value if v is not None)
        else:
            match_null = ' OR {0} IS NULL'
        return cls('({0} NOT IN %s' + match_null + ')', value)

    @classmethod
    def inselect(cls, inselect_query):
        return cls('{0} IN (' + inselect_query + ')', )

    @classmethod
    def not_inselect(cls, inselect_query):
        return cls('{0} NOT IN (' + inselect_query + ')', )

    def _merge(self, other, sep):
        cls = type(self)
        q = '(({0}) {1} ({2}))'.format(self._query, sep, other._query)
        a = self._values + other._values
        return cls(q, *a)

    def __and__(self, other):
        return self._merge(other, 'AND')

    def __or__(self, other):
        return self._merge(other, 'OR')


# ----------------------------------------------------------------------------
# Branchs & ports functions

def _branch_key(branch):
    if branch in BRANCH_ALIASES:
        return BRANCH_ALIASES[branch]
    if '.' in branch:
        return int(re.sub(r'\D', '', branch))
    return int(branch.rpartition('-')[-1])   # XXX branch-tpl


def branch_is_running(branch):
    pidfile = 'openerp-{0}.pid'.format(branch)
    return os.path.isfile(os.path.join(PID_DIR, pidfile))


def get_port(branch):
    if TEST and not LOCAL and USER != 'odoo':
        return {
            'master': 7069,
            'demo': 7169,
            'trial': 7269,
            'accounts': 7369,
            'apps': 7469,
            'iap': 7569,
        }[USER]
    if branch == 'master':
        return 9999
    k = _branch_key(branch)
    if k < 100:     # <= 11.0
        return int(8069 + 100 * k)  # XXX port-computation
    return 42000 + k


def get_last_branch(all_=False):
    branches = list(list_branches(all_))
    if branches:
        return branches[-1]
    raise RuntimeError('No running branch found')


def branch_exists(branch):
    return os.path.isdir(os.path.join(ODOO_SRC, branch))


def list_branches(all_=False):
    candidates = sorted((e for e in os.listdir(ODOO_SRC) if not e.startswith('.')), key=_branch_key)
    for c in candidates:
        if all_ or branch_is_running(c):
            yield c


def get_branch(port):
    for b in list_branches(True):
        if get_port(b) == port:
            return b
    return None

# ----------------------------------------------------------------------------


def _min_cron_round(cr, port):
    cr.execute("""SELECT coalesce(min(cron_round), 0)
                    FROM databases
                   WHERE mode IN %s
                     AND coalesce(status, '') NOT IN %s
                     AND port = %s
               """, (ACTIVE_MODES, DISABLED_STATUSES, port))
    return cr.fetchone()[0]


def _next_cron(port):
    """
        Returns a 3-tuple (dbname, rest, total_count) which are:
            - the name of the database on which the cron must run
            - count of database to process in this round
            - total count of databases for this series (port)
    """
    with cursor() as cr:
        cr.execute("""
            WITH candidates AS (
                SELECT name, coalesce(cron_round, 0) as cron_round
                  FROM databases
                 WHERE mode IN %s
                   AND coalesce(status, '') NOT IN %s
                   AND port = %s
            )
            SELECT min(name) dbname, count(1) inround, (SELECT count(1) FROM candidates) as total
              FROM candidates
             WHERE cron_round = (SELECT min(cron_round) FROM candidates)
        """, (ACTIVE_MODES, DISABLED_STATUSES, port))
        return cr.fetchone()


def add(name, port=None, expire_in=None, create_date="now() at time zone 'UTC'", secure=True, **kw):
    if port is None:
        port = get_port(get_last_branch())

    with cursor() as cr:
        expire_date = kw.pop('expire_date', None)
        if expire_date is None:
            expire_date = 'NULL' if expire_in is None else "now() at time zone 'UTC' + interval '%s'" % expire_in
        mode = kw.get('mode')
        if mode is None:
            mode = 'paid' if expire_in is None else 'trial'

        kw.setdefault('cron_round', _min_cron_round(cr, port))

        kw.update({
            'name': name,
            'port': port,
            'mode': mode,
            'secure': secure,
        })
        kw.pop('id', None)

        if 'email_daily_limit' not in kw and mode == 'paid':
            kw['email_daily_limit'] = EMAIL_DAILY_LIMIT_PAID

        if 'duplicate_limit' not in kw and mode == 'paid':
            kw['duplicate_limit'] = DUPLICATE_LIMIT_PAID

        _logger.info('new db => %r', kw)
        data = kw.items()
        cols = ", ".join(['create_date', 'expire_date'] + [d[0] for d in data])
        values = ", ".join([sql.quote_dt(create_date), sql.quote_dt(expire_date)] + ["%s"] * len(data))

        query = "INSERT INTO databases({0}) VALUES ({1}) RETURNING id".format(cols, values)
        cr.execute(query, list(map(itemgetter(1), data)))

        return cr.fetchone()[0]


def remove_database(dbname):
    with cursor() as cr:
        cr.execute("DELETE FROM databases WHERE name=%s", (dbname,))
        return True


def set_hostname(cr, port, hostname, slaves=(), cname_list=()):
    kw = {'port': port} if TEST and not LOCAL else {}     # ports are hardcoded in TEST
    db = get('', mode='hostname', _order='create_date', **kw)
    if not db:
        db_id = add('', port=port, mode='hostname')
    else:
        if db.port != port:
            update(db, port=port)
        db_id = db.id
    hosts = set([hostname.lower()] + [h.lower() for h in slaves + cname_list])

    cr.execute("SELECT name FROM domains WHERE db_id=%s OR name IN %s", [db_id, tuple(hosts)])
    existing_hosts = set(map(itemgetter(0), cr.fetchall()))

    missings = hosts - existing_hosts
    for m in missings:
        cr.execute("INSERT INTO domains (db_id, name) VALUES(%s, %s)", [db_id, m])

    cr.execute("DELETE FROM domains WHERE db_id=%s AND name NOT IN %s", [db_id, tuple(hosts)])


def add_host(cr, db, host):
    if host.endswith(SAAS_VALID_SUFFIXES):
        return None
    assert isnamedtuple(db)
    host = host.lower()
    cr.execute("SELECT id FROM domains WHERE db_id=%s AND name=%s", (db.id, host))
    [hid] = cr.fetchone() or [None]
    if not hid:
        _logger.info('add host %s for db %s', host, db.name)
        cr.execute("INSERT INTO domains(db_id, name) VALUES(%s, %s) RETURNING id", (db.id, host))
        hid = cr.fetchone()[0]
    return hid


# used by saas_worker routes (/domain-names/*) to set custom domains to databases
def get_hostnames(dbname):
    with cursor() as cr:
        db = get(dbname)
        cr.execute("SELECT name FROM domains WHERE db_id=%s", (db.id,))
        return [record[0] for record in cr.fetchall()]

def remove_hostnames(dbname):
    with cursor() as cr:
        cr.execute("DELETE FROM domains d USING databases db where d.db_id = db.id and db.name = %s", [dbname])

def add_hostname(dbname, host):
    with cursor() as cr:
        db = get(dbname)
        return add_host(cr, db, host)


def remove_hostname(dbname, host):
    with cursor() as cr:
        db = get(dbname)
        cr.execute("DELETE FROM domains WHERE db_id=%s AND name=%s", (db.id, host.lower()))
        return True
# end of saas_worker (/domain-names/*) routes utilities


def _link_db(cr, mig):
    assert isnamedtuple(mig)
    if isnamedtuple(mig.db_id):
        return mig
    return mig._replace(
        db_id=next(_each(cr, id=mig.db_id)),
        copy_id=next(_each(cr, id=mig.copy_id)) if mig.copy_id else None,
    )


def add_migration(db, create_date="now() at time zone 'UTC'", state='queued', **kw):
    assert isnamedtuple(db)
    if has_pending_migrations(db, kw.get('mode') in ('rr-preflight', 'rr-preflight-muted')):
        return False
    if db.name.startswith(EDU_PREFIX):
        return False
    with cursor() as cr:
        kw.update({
            'db_id': db.id,
            'state': state,
        })
        kw.pop('id', None)

        _logger.info('add %s migration for db %s', state, db.name)

        data = kw.items()
        cols = ", ".join(['create_date'] + list(map(itemgetter(0), data)))
        values = ", ".join([sql.quote_dt(create_date)] + ["%s"] * len(data))

        query = 'INSERT INTO migrations({0}) VALUES ({1}) RETURNING *'.format(cols, values)
        cr.execute(query, list(map(itemgetter(1), data)))

        return _link_db(cr, cr.fetchone())


def update_migration(mig, **kw):
    assert isnamedtuple(mig)
    q = []
    d = ()
    for k, v in kw.items():
        if k == 'id':
            raise KeyError('Cannot update id')
        if k in DT_COLUMNS:
            q += ['%s=%s' % (k, sql.quote_dt(v))]
        else:
            q += ['%s=%%s' % (k,)]
            d += (v,)

    d += (mig.id,)

    query = 'UPDATE migrations SET '
    query += ', '.join(q)
    query += ' WHERE id=%s'
    with cursor() as cr:
        mig = _link_db(cr, mig)
        _logger.info('update migration #%d for db %s => %r', mig.id, mig.db_id.name, kw)
        cr.execute(query, d)
        cr.execute('SELECT * FROM migrations WHERE id=%s', [mig.id])
        return _link_db(cr, cr.fetchone())

def migration_prepare_notify(dbname, mode, target_version, state):
    """
    called after an interactive migration
    - update migration record.
    - notify accounts if this condition is true:
        - we are at the end of a succesful interactive migration and a queued
          migration has already been attempted and failed.
    """
    db = get(dbname)
    args = {'state': state}
    if mode == 'copy':
        args['copy_id'] = db.id

    # if an existing failed migration request is found, update the record
    mig = None
    re_notify = False
    for m in get_migrations(db):
        # FIXME do a proper SQL query
        if m.target_version == target_version and m.mode == mode and m.state == 'failed':
            mig = m
            if state != mig.state:
                re_notify = True
                mig = update_migration(mig, **args)
            break

    # conditionally notify accounts
    if re_notify:
        migration_notify(mig)


def migration_notify(mig):
    notify_url = WEBSITE + '/odoo-enterprise/migration-notify'
    notify_kw = {
        'db': mig.db_id.uuid,
        'to': mig.target_version,
        'mode': mig.mode,
        'state': mig.state,
        'email_to': mig.email,
        'db_url': '%s://%s.%s' % (SCHEME, mig.db_id.name, SAAS_BASE_URL),
    }
    if mig.copy_id:
        notify_kw['copy'] = mig.copy_id.uuid
    requests.post(notify_url, notify_kw, verify=not TEST, auth=SaasAuth(), timeout=TIMEOUT)


def migration_delegate(cr, mig):
    """ Delegate to the upgrade platform. """
    upgrade_api = get_upgrade_api(mig.db_id, mig.source_version, pg_cnx, pg_env)
    if mig.mode == 'self':
        # lock prod before upload
        # TODO?: compute interval based on last test migration duration? or leave 4h?
        interval = '2 hours' if mig.actuator == 'rr' else '12 hours'
        mig = update_migration(mig, date_limit="now() at time zone 'UTC' + interval '%s'" % interval)
        update(mig.db_id, status='maintenance_mig')
    try:
        _update_module_list(mig.db_id)
        request_info = upgrade_api.new_request(mig.db_id.name, mig.target_version, mig.actuator, mig.mode, mig.email)
    except UpgradeError as e:
        _migration_abort(mig, e.args[0])
        return

    args = {
        'state': 'queued',
        'request_id': request_info['request'],
        'request_key': request_info['key'],
    }
    mig = update_migration(mig, **args)


def _each(cr, _order=None, _limit=None, **kw):
    query = 'SELECT * FROM databases'
    where = []
    data = ()

    for k, v in kw.items():
        if k not in COLUMNS:
            raise KeyError('Invalid column: %r. %r' % (k, COLUMNS))
        if not isinstance(v, sql):
            v = sql.eq(v)

        w, d = v.bind(k)
        where += [w]
        data += d

    if where:
        query += ' WHERE ' + ' AND '.join(where)

    if _order:
        query += ' ORDER BY ' + _order   # FIXME sqli

    if _limit:
        query += ' LIMIT %s'
        data += (_limit,)

    cr.execute(query, data)
    for row in cr.fetchall():
        if row.source_uuid:
            try:
                row = row._replace(source_uuid=next(_each(cr, uuid=row.source_uuid)))
            except StopIteration:
                row = row._replace(source_uuid=None)
        yield row


def list_(**kw):
    with cursor() as cr:
        for db in _each(cr, **kw):
            yield db


def get(dbname, **kw):
    """
    Reads db info by its name and returns a named tuple of it's record
    :param dbname: str
    :param kw: filter, key values to restrict records when searching. Usually used for 'mode' queries
               if 'mode' is not in kw, it will be added to filter active databases: trial, drop and paid.
    :return: Named tuple of the record
    """
    with cursor() as cr:
        try:
            kw['name'] = dbname
            if 'mode' not in kw:
                kw['mode'] = sql.in_('trial', 'paid', 'drop')
            return next(_each(cr, **kw))
        except StopIteration:
            return None


def get_duplicates_count(db):
    """
    Get the number of databases with the same source_uuid as the given database uuid.
    The duplicates are counted by mode. Each mode is counted separately and ponderated by
    a fixed factor:
    - 1.0 for mode 'trial'
    - 0.25 for mode 'drop'
    - 0.1 for mode 'paid'

    :param db: named tuple of the database
    :return: float, the number of duplicates
    """

    dups_name = get_duplicates(db)
    if not dups_name:
        return 0

    with cursor() as cr:
        cr.execute("""
            SELECT mode, count(id) FROM databases WHERE name IN %s GROUP BY mode;
        """, [tuple(dups_name)])
        duplicate_count = cr.fetchall()

    ponderation_dict = {
        'trial': 1.0,
        'drop': 0.25,
        'paid': 0.1,
    }

    # Use ponderation dict to compute the sum. Use get if mode has unexpected value
    return sum(ponderation_dict.get(dup[0], 0) * dup[1] for dup in duplicate_count)


def get_duplicates(db):
    with cursor() as cr:
        cr.execute("""
            WITH RECURSIVE dups (uuid, source_uuid, name) AS (
                SELECT uuid, source_uuid, name FROM databases WHERE uuid = %s
              UNION
                SELECT db.uuid, db.source_uuid, db.name
                FROM dups dup, databases db
                WHERE db.uuid = dup.source_uuid OR db.source_uuid = dup.uuid
            )
            SELECT name FROM dups WHERE source_uuid IS NOT NULL AND uuid != %s ORDER BY name;
        """, [db.uuid, db.uuid])
        dups = [dup[0] for dup in cr.fetchall()]
    return dups


def get_by_uuid(uuid):
    with cursor() as cr:
        try:
            return next(_each(cr, uuid=uuid))
        except StopIteration:
            return None


def get_hosts(db):
    assert isnamedtuple(db)
    with cursor() as cr:
        cr.execute("SELECT name FROM domains WHERE db_id=%s", [db.id])
        domains = [h[0] for h in cr.fetchall()]
    return [db.name + s for s in SAAS_VALID_SUFFIXES] + domains


def get_migrations(db):
    assert isnamedtuple(db)
    with cursor() as cr:
        cr.execute("SELECT * FROM migrations WHERE db_id=%s order by id desc", (db.id,))
        for m in cr.fetchall():
            yield _link_db(cr, m)


def get_last_upgrade(db):
    assert isnamedtuple(db)
    with cursor() as cr:
        cr.execute("SELECT * FROM migrations WHERE db_id=%s order by id desc limit 1", (db.id,))
        if not cr.rowcount:
            return None
        m = cr.fetchone()
        return _link_db(cr, m)


def get_by_host(host):
    """
    gets dbname from domain
    :param host: string, domain (e.g. 'example.com')
    :return: dbname or None
    """
    assert isinstance(host, (type(''), type(u'')))
    host = host.lower()
    with cursor() as cr:
        db = None
        if host.endswith(SAAS_VALID_SUFFIXES):
            dbname = host.partition('.')[0]
            cr.execute("SELECT id FROM databases WHERE name=%s", [dbname])
            [db] = cr.fetchone() or [None]
        if not db:
            cr.execute("SELECT db_id FROM domains WHERE name=%s", [host])
            [db] = cr.fetchone() or [None]
        if not db:
            return None
        return next(_each(cr, id=db))


def get_stats(db, date_from=None, date_to=None):
    with cursor() as cr:
        where_params = (db,)
        where_conditions = ['d.name = %s'] # everything put in this list will be concatenated with " AND "

        if date_from:
            where_conditions.append('s.date >= %s')
            where_params = where_params + (date_from,)

        if date_to:
            where_conditions.append('s.date <= %s')
            where_params = where_params + (date_to,)

        selected_fields = ['s."{0}"'.format(f) for f in DB_STAT_FIELDS]
        query = (
            'SELECT d.name, {selected_fields} '
            'FROM db_stats s '
            'JOIN databases d '
            '    ON d.id = s.db_id '
            'WHERE {where_conditions} '
            'ORDER BY s."date" '
        ).format(
            selected_fields=', '.join(selected_fields),
            where_conditions=' AND '.join(where_conditions),
        )
        cr.execute(query, where_params)
        return cr.fetchall()


def get_domains(db):
    with cursor() as cr:
        where_params = (db,)
        where_conditions = ['db.name = %s'] # everything put in this list will be concatenated with "AND "

        selected_fields = ['d."{0}"'.format(f) for f in DB_DOMAIN_FIELDS]
        query = (
            'SELECT db.name db_name, {selected_fields} '
            'FROM domains d '
            'JOIN databases db ON db.id = d.db_id '
            'WHERE {where_conditions} '
        ).format(
            selected_fields=', '.join(selected_fields),
            where_conditions=' AND '.join(where_conditions),
        )
        cr.execute(query, where_params)
        return cr.fetchall()


def get_stats_total_email(db, date=None, date_included=True):
    with cursor() as cr:
        where_params = (db,)
        where_conditions = ['d.name = %s'] # everything put in this list will be concatenated with " AND "

        if date:
            where_conditions.append('date %s %%s' % ('<=' if date_included else '<',))
            where_params = where_params + (date,)

        query = (
            'SELECT coalesce(sum(s.email_count), 0) total_email, '
                   'coalesce(sum(s.bounce_count), 0) total_bounce '
            'FROM db_stats s '
            'JOIN databases d ON d.id = s.db_id '
            'WHERE {where_conditions} '
        ).format(
            where_conditions=' AND '.join(where_conditions),
        )
        cr.execute(query, where_params)
        return cr.fetchone()


def trial_list():
    with cursor() as cr:
        for r in _each(cr, mode='trial', _order='name'):
            yield r.name


def get_expired_dbs():
    """
    Build a list of trial dbs to be dropped.
    """
    with cursor() as cr:
        filters = [
            dict(mode='trial', status=sql.neq('maintenance', wildcard=True)),
            dict(mode='drop'),
        ]
        for f in filters:
            for r in _each(cr, expire_date=sql("{0} < now() at time zone 'UTC'"), **f):
                yield r


def get_ghost_db_names():
    """"
    Retrieve DB names in meta but not in postgres, created at least 7 days ago.
    """
    with cursor() as cr:
        query = """
            SELECT m.name
            FROM databases m
            LEFT JOIN pg_database d ON d.datname = m.name
            WHERE d.datname IS NULL
            AND m.mode in ('trial', 'drop')
            AND m.status not in ('frozen', 'abuse', 'to_activate')
            AND m.create_date < now() at time zone 'UTC' - interval '7 days'
        """
        cr.execute(query)
        return [d[0] for d in cr.fetchall()]


def get_failed_trial_creation():
    """
    Get databases whose creation has failed, mainly due to service restart.
    """
    with cursor() as cr:
        query = """
            SELECT name
            FROM databases
            WHERE mode = 'trial'
            AND status = 'maintenance'
            AND expire_date < now() at time zone 'UTC'
            AND name like '_%'
        """
        cr.execute(query)
        return [d[0] for d in cr.fetchall()]


def spare_list():
    return list(list_(mode='spare'))


def spare_use():
    spares = spare_list()
    with cursor() as cr:
        for spare in spares:
            try:
                cr.execute('SELECT * FROM databases WHERE name=%s FOR UPDATE NOWAIT', (spare.name,))
                cr.execute('DELETE FROM databases WHERE name=%s', (spare.name,))
            except psycopg2.OperationalError:
                pass
            else:
                return spare.name
    raise Exception("No spare database available")


def set_expiry_in(dbname, interval, set_mode=None, set_status=None, set_extra_apps=None):
    db = get(dbname)
    kw = {'expire_date': 'NULL' if interval is None else "now() at time zone 'UTC' + interval '%s'" % interval}
    if set_mode:
        kw['mode'] = set_mode
        if set_mode == 'paid' and db.email_daily_limit < EMAIL_DAILY_LIMIT_PAID:
            kw['email_daily_limit'] = EMAIL_DAILY_LIMIT_PAID
        if set_mode == 'paid' and db.duplicate_limit < DUPLICATE_LIMIT_PAID:
            kw['duplicate_limit'] = DUPLICATE_LIMIT_PAID
    if set_status:
        kw['status'] = set_status
    if set_extra_apps:
        kw['extra_apps'] = set_extra_apps
    update(db, **kw)


def update(db, **kw):
    assert isnamedtuple(db)
    assert kw

    # pre-check keys
    for k in kw:
        if k not in COLUMNS:
            raise KeyError('Invalid column: %r. %r' % (k, COLUMNS))

        if k == 'id':
            raise KeyError('Cannot update id')

    _logger.info('update db %s => %r', db.name, kw)

    q = []
    d = ()

    with cursor() as cr:
        if 'cron_round' not in kw:
            port = int(kw.get('port', db.port))
            if port != db.port:
                # port change => match min value
                kw['cron_round'] = _min_cron_round(cr, port)

        for k, v in kw.items():
            if k in DT_COLUMNS:
                q += ['%s=%s' % (k, sql.quote_dt(v))]
            else:
                q += ['%s=%%s' % (k,)]
                d += (v,)

        d += (db.id,)

        query = 'UPDATE databases SET '
        query += ', '.join(q)
        query += ' WHERE id=%s'
        cr.execute(query, d)


def increment(db, **kw):
    """
    Increment the given fields by their associated value.
    For example, `metabase.increment(db, email_daily_count=3)`
    would increment the email daily count by 3.

    With `metabase.update`, you need to read the value, increment it on Python side and write it.
    Therefore there is a risk that the value was changed in the meantime.

    With `metabase.increment`, the incrementation is atomic.
    The incrementation will hang until the columns are unlocked by other transactions,
    consider the last updated value, and increment it with SQL.
    """
    assert isnamedtuple(db)
    assert kw

    # pre-check keys
    for k in kw:
        if k not in COLUMNS:
            raise KeyError('Invalid column: %r. %r' % (k, COLUMNS))

        if k in ('id', 'cron_round') or k in DT_COLUMNS:
            raise KeyError('Cannot increment %s' % (k,))

    _logger.info('increment db %s => %r', db.name, kw)

    q = []
    d = ()

    with cursor() as cr:
        for k, v in kw.items():
            q += ['{0}={0} + %s'.format(k)]
            d += (v,)

        d += (db.id,)

        query = 'UPDATE databases SET '
        query += ', '.join(q)
        query += ' WHERE id=%s'
        cr.execute(query, d)


def _update_table_from_json(db_id, table, fields, json):
    if json.get('fields', []) and json.get('data', []):
        with cursor() as cr:
            # Transform the structure of (M records, N fields) from:
            # {fields: ['field1', ..., 'fieldN'], data: [['val1.1', ..., 'val1.N'], ..., ['valM.1', ..., 'valM.N']]}
            # where the fields may be ordered differently than DB_STAT_FIELDS
            # to a list of records that matches our DB_STAT_FIELDS
            # Missing fields will be silently ignored
            fields = tuple(fields & set(json['fields']))
            fields2idx = {f: json['fields'].index(f) for f in fields}
            values = [(db_id,) + tuple(record[fields2idx[f]] for f in fields) for record in json['data']]
            fields = ('db_id',) + fields
            # build the query to insert all these formatted records
            query = 'INSERT INTO {0} ({1}) VALUES {2}'.format(
                table,
                ','.join('"{}"'.format(f) for f in fields),
                ','.join(['%s'] * len(values)),
            )
            cr.execute(query, values)


@contextlib.contextmanager
def disable_email_limit(dbname):
    db = get(dbname)
    count, ext_count, total = db.email_daily_count, db.ext_email_daily_count, db.email_total_count
    update(db, email_daily_limit=0x7FFFFFFF) # max signed int 32
    try:
        yield
    finally:
        update(db, email_daily_count=count, ext_email_daily_count=ext_count, email_daily_limit=50, email_total_count=total)

def _get_size(dbname):
    """
    Copied from util.py to avoid cyclic imports.
    """
    # Dump size
    bakdir = os.path.expanduser('~/bakdb')
    backup_size = 0
    try:
        for backup_file in os.listdir(bakdir):
            # backups files have the file name pattern:
            # <dname>.<uuid>.<date>.sql.gz
            if backup_file.startswith(dbname + '.') and backup_file.endswith('.sql.gz'):
                backup_size = os.path.getsize(os.path.join(bakdir, backup_file))
                break
    except OSError:
        pass

    # Filestore size
    fs_path = os.path.join(os.path.expanduser('~/filestore'), dbname)
    fs_size = sum(os.path.getsize(os.path.join(dirpath, filename))
                  for dirpath, _, filenames in os.walk(fs_path)
                  for filename in filenames)

    return(backup_size, fs_size)

class MigrationError(Exception):
    pass


_EXCLUDED = None


def pg_call(cmd):
    # TODO uncomment logger call once the logger feature is merged
    # _logger.info('pg_call: %s', cmd)
    prog, _, tail = cmd.lstrip().partition(' ')
    cmd = '%s %s %s' % (prog, pg_cnx, tail)
    return subprocess.call(cmd, shell=True, env=pg_env)

def drop_conn(name):
    pid_col = "pid" if _pg_server_version >= 90200 else "procpid"
    query = (
        "SELECT pg_terminate_backend({0})"
        "  FROM pg_stat_activity"
        " WHERE datname = '{1}' AND {0} != pg_backend_pid();"
    ).format(pid_col, name)
    pg_call('psql postgres -c "{}" >/dev/null'.format(query))


def is_migration_forbid(db, version):
    global _EXCLUDED
    assert isnamedtuple(db)
    if _EXCLUDED is None:
        x = {}
        p = '/home/odoo/migration-exclude'
        if not os.path.exists(p):
            return False
        for s in os.listdir(p):
            with open(os.path.join(p, s)) as f:
                x[get_port(s)] = frozenset(dd for d in f.readlines() for dd in [d.strip()]
                                           if dd and not dd.startswith('#'))
        _EXCLUDED = x

    port = get_port(version)
    return any(db.name in l for p, l in _EXCLUDED.items() if db.port < p <= port)


def migrate(dbname, to, verbose=False, force=False, keep_failed=False, copy=False,
            inplace=False, update_list=True, oe_meta_command='migrate', upgrade_path=None):
    verbose = int(verbose)
    if verbose >= 1:
        _logger.setLevel(logging.DEBUG)
    if not TEST and USER != 'odoo':
        raise MigrationError('Must be run as odoo')

    if dbname == 'meta':
        raise MigrationError('Nice try')

    db = get(dbname)
    if not db:
        raise MigrationError('Unknow database %r' % (dbname,))

    # TODO check db exists in pg_databases
    if not branch_exists(to):
        raise MigrationError('Unknow branch %r' % (to,))

    if not copy and is_migration_forbid(db, to):
        raise MigrationError('Database %r excluded from migration to %s' % (dbname, to))

    if not force and not branch_is_running(to):
        raise MigrationError('Branch %s is not running' % (to,))

    dest_port = get_port(to)
    if not force and db.port >= dest_port:
        raise MigrationError('Database %r is newer than %s' % (dbname, to))

    if db.status and db.status.startswith('maintenance'):
        raise MigrationError('Database %r is already in maintenance' % (dbname,))

    migs = list(m for m in get_migrations(db) if dest_port == get_port(m.target_version))
    if not copy and any(m.state == 'queued' and m.mode == 'copy' for m in migs):
        raise MigrationError('A copy migration is already queued for database %r' % (dbname,))

    if copy and inplace:
        raise MigrationError("--copy cannot be used in conjonction with --inplace")

    mode = ['self', 'copy'][bool(copy)]
    proctitle = 'oe-meta {command}: {db} -> {version} ({mode})'.format(
        command=oe_meta_command, mode=mode, db=dbname, version=to)
    setproctitle(proctitle)
    _logger.info('execute %s', proctitle)

    # Used by some migration scripts for SaaS-specific modules
    mig_env = dict(os.environ, OE_SAAS_MIGRATION='1')
    FS = os.path.expanduser('~/filestore')

    def update_meta(**kw):
        if verbose >= 1:
            upd = ', '.join("{0}={1}".format(k, "'%s'" % v if isinstance(v, (type(''), type(u''))) else v)
                            for k, v in kw.items())
            # actually update via db.id
            _logger.debug("@ UPDATE databases SET {0} WHERE name='{1}'".format(upd, dbname))
        update(db, **kw)

    def call(cmd, env=None):
        assert isinstance(cmd, str)
        _logger.debug("+ %s", cmd)
        return subprocess.call(cmd, shell=True, env=env)

    def dropdb(name):
        drop_conn(name)
        pg_call('dropdb "{0}"'.format(name))
        call('rm -rf {0}/{1}'.format(FS, name))

    def renamedb(from_, to, fs=True):
        drop_conn(from_)
        drop_conn(to)
        pg_call("""psql -qc 'ALTER DATABASE "{0}" RENAME TO "{1}"' postgres""".format(from_, to))
        if fs:
            call('mv {0}/{1} {0}/{2}'.format(FS, from_, to))

    update_meta(status='maintenance_mig')
    time.sleep(3)
    drop_conn(dbname)

    if update_list:
        _update_module_list(db)

    t = int(time.time())
    if inplace:
        migdb = dbname
    else:
        migdb = '_migrate_%s_%s' % (dbname, t)

        rc = pg_call('createdb --template="{0}" "{1}"'.format(dbname, migdb))
        if rc:
            update_meta(status=db.status)
            raise MigrationError('Failed to duplicate database %r' % (dbname,))

        db_fs = os.path.join(FS, dbname)
        if not os.path.exists(db_fs):
            os.mkdir(db_fs)
        if copy:
            call('cp -al {0}/{1} {0}/{2}'.format(FS, dbname, migdb))
        else:
            # for non-copy upgrades, starts with an empty filestore that will be rsynced into the
            # origin one once the upgrade succeed.
            mig_fs = os.path.join(FS, migdb)
            if not os.path.exists(mig_fs):
                os.mkdir(mig_fs)

    if copy:
        update_meta(status=db.status)

    # NOTE modules marked to be remove are still loaded by openerp and usable by customers.
    #      "to install" are not. These states are mean to be temporary during update.
    #      Reset state of these module to avoid data deletion.
    pg_call('''psql -qc "update ir_module_module set state=case when state='to install' then 'uninstalled' when state in ('to remove', 'to upgrade') then 'installed' else state end" {0}'''.format(migdb))
    pg_call('''psql -qc "update ir_module_module set latest_version=NULL where state != 'installed'" {0}'''.format(migdb))

    # upgrade to new version
    loglevel = ['ERROR', 'ERROR', 'INFO', 'DEBUG'][min(verbose, 3)]
    loghandlers = [':' + loglevel]
    if verbose <= 1:
        loghandlers += [
            'odoo.addons.base.maintenance:WARNING',
            # hide missing dep errors (we handle it)
            'odoo.modules.graph:CRITICAL',
            # display warning about invalid custom views
            'odoo.modules.loading:WARNING',
        ]
    if verbose >= 1:
        loghandlers += ['odoo.addons.base.maintenance:INFO', 'odoo.modules.migration:INFO']
    if verbose >= 2:
        loghandlers += ['odoo.modules.loading:DEBUG']

    loghandlers = ' --log-handler='.join([''] + loghandlers)

    upgrade_cmd = '~/bin/openerp-{0} -d {1} -u all --stop-after-init {2} --logfile= --xmlrpc-port=0 --workers=0'.format(to, migdb, loghandlers)
    if upgrade_path:
        upgrade_cmd += ' --upgrade-path={0}'.format(upgrade_path)
    rc = call(upgrade_cmd, env=mig_env)
    if rc:
        if not inplace:
            if keep_failed:
                failed = '_failed_migration_%s_%s' % (dbname, t)
                renamedb(migdb, failed)
            else:
                dropdb(migdb)

            if not copy:
                update_meta(status=db.status)

        raise MigrationError('Failed to migrate database %r' % (dbname,))

    if not copy and not inplace:
        tmpdb = '_tmp_%s_%s' % (dbname, t)
        renamedb(dbname, tmpdb, fs=False)
        renamedb(migdb, dbname, fs=False)
        call("rsync --archive --quiet {0}/{1}/ {0}/{2}/".format(FS, migdb, dbname))
        call("rm -rf {}/{}".format(FS, migdb))
        dropdb(tmpdb)

    if not copy:
        if TEST:
            update_meta(status=db.status)
        else:
            update_meta(status=db.status, port=dest_port)
    else:
        tto = to.replace('-', '').split('.')[0]
        dup_name = '{0}-test-{1}'.format(dbname, tto)
        res = requests.get('{0}/saas_master/hostof?dbname={1}'.format(MASTER, dup_name), auth=SaasAuth(), timeout=MASTER_TIMEOUT)
        if res.content:
            dup_name = '{0}-{1}'.format(dup_name, uuid.uuid4().hex[:5])
        hostname = '%s.%s' % (dup_name, SAAS_BASE_URL)
        new_uuid = str(uuid.uuid1())
        migration_neuter(migdb, new_uuid, hostname)
        renamedb(migdb, dup_name)
        copy_id = add(
            dup_name, dest_port, '15 days',
            uuid=new_uuid, source_uuid=db.uuid,
            extra_apps=True
        )

        with open(os.path.expanduser('~/config/config.json')) as fp:
            config = json.load(fp)
        host = config['me']['name']

        resp = requests.post(MASTER + '/saas_master/create-dns', {
            'dbname': dup_name,
            'host': host,
        }, verify=not TEST, auth=SaasAuth(), timeout=MASTER_TIMEOUT)
        resp.raise_for_status()

        dbname = dup_name

    _force_ping(dbname, to)

    if oe_meta_command != 'queue-migrate':
        migration_prepare_notify(dbname, mode, to, 'done')

    return {
        'db_id': db.id,
        'copy_id': copy_id if copy else None,
    }

def migration_neuter(dbname, new_uuid, hostname):
    raise Exception("migration_neuter() is deprecated and should be used anymore.")

@graceful_stop
def auto_migrate(from_, to):
    from_port = get_port(from_)
    if to is None:
        to = get_last_branch()
    to_port = get_port(to)
    if from_port >= to_port:
        raise MigrationError('Cannot auto-migrate from %s to %s' % (from_, to))
    with cursor() as cr:
        for db in _each(
            cr,
            id=sql.inselect(cr.mogrify("""
            SELECT m.db_id
              FROM migrations m
              JOIN databases d ON d.id = m.db_id
             WHERE m.source_version=%s
               AND m.target_version=%s
               AND m.mode IN ('rr-preflight', 'rr-preflight-muted')
               AND m.state IN ('rolling_release_available', 'rolling_release_warning')
               AND (d.rr_locked_until IS NULL OR d.rr_locked_until < CURRENT_DATE)
               AND m.create_date - CURRENT_DATE < interval '1 month'
               AND NOT EXISTS (SELECT 1
                                 FROM migrations m2
                                WHERE m2.source_version=m.source_version
                                  AND m2.db_id=m.db_id
                                  AND m2.mode in ('rr-forced', 'self')
                                  AND m2.state in ('queued', 'done')
                              )
               AND NOT EXISTS (SELECT 1
                                 FROM migrations m3
                                WHERE m3.source_version=m.source_version
                                  AND m3.target_version=m.target_version
                                  AND m3.db_id=m.db_id
                                  AND m3.state='failed'
                                  AND m3.id>m.id
                              )
          GROUP BY 1
            """, [from_, to]).decode()),
            mode=sql.in_('paid', 'trial'),
            port=from_port,
            status=sql.in_('open', 'pending', 'activated', '', None)
        ):
            if not db.name.startswith(EDU_PREFIX):
                add_migration(db, actuator="meta", mode="rr-forced", source_version=from_, target_version=to)


def auto_migrate_db(dbname, target_version, mode=False):
    if dbname == 'meta':
        raise MigrationError('Nice try')
    if dbname.startswith(EDU_PREFIX):
        raise MigrationError('No migration for EDU databases')
    db = get(dbname)
    if db:
        if not mode:
            mode = 'rr-preflight'
        if mode not in ('rr-preflight', 'rr-preflight-muted', 'rr-forced', 'self', 'copy'):
            raise MigrationError('Invalid mode: %s, available modes are rr-preflight, rr-preflight-mited, rr-forced, self and copy' % mode)
        if db.rr_locked_until and db.rr_locked_until >= datetime.datetime.now() and mode not in ('rr-preflight', 'rr-preflight-muted'):
            raise MigrationError('Database migrations are locked until %s' % db.rr_locked_until)
        version = get_branch(db.port)
        if db.port >= get_port(target_version):
            raise MigrationError('Cannot auto-migrate from %s to %s' % (version, target_version))
        state_filestore = "pending" if mode in ("self", "copy") else None
        add_migration(
            db,
            actuator="meta",
            mode=mode,
            source_version=version,
            target_version=target_version,
            state_filestore=state_filestore,
        )


def queue_migrate(migrations_to_handle):
    """"""
    for _ in range(max(migrations_to_handle, 1)):
        rr_forced_process_queue()
        process_delegate_queue()
        if not is_queued_migration():
            break
    update_queue()
    prepare_filestore()
    remote_restore_init()
    local_restore_queue()
    finalize_prod_migration()
    process_cleanup()
    rr_preflight_process()
    unlock_maintenance_mig()


def is_queued_migration():
    query = """
        SELECT 1
          FROM migrations
         WHERE state = 'queued'
           AND request_id IS NULL
         LIMIT 1
    """
    with cursor() as cr:
        cr.execute(query)
        return bool(cr.rowcount)


def has_pending_migrations(db, include_preflight=False):
    assert isnamedtuple(db)
    with cursor() as cr:
        query = """
            SELECT 1
              FROM migrations
             WHERE db_id=%s
               AND state='queued'
        """
        if not include_preflight:
            query += " AND NOT mode in ('rr-preflight', 'rr-preflight-muted')"
        cr.execute(query, (db.id,))
        return bool(cr.rowcount)


@log_call
def process_delegate_queue():
    #TODO: enforce locking by using FOR UPDATE SKIP LOCKED once all clusters are on 9.5+
    # Handle production(=self) before test(=copy) then rolling-release
    # skip rr-forced migration, handled with another part of the cron job
    query = """
        SELECT *
          FROM migrations
         WHERE state = %s
           AND request_id IS NULL
           AND mode !='rr-forced'
      ORDER BY CASE mode
                 WHEN 'self' THEN 1
                 WHEN 'copy' THEN 2
                 ELSE 3
               END
         LIMIT 1
    """
    with cursor() as cr:
        s_queued = 'queued'
        cr.execute(query, [s_queued])
        for mig in cr.fetchall():
            migration_delegate(cr, _link_db(cr, mig))

@log_call
def update_queue():
    with cursor() as cr:
        cr.execute("SELECT * FROM migrations WHERE state = 'queued' AND request_id IS NOT NULL")
        for m in cr.fetchall():
            m = _link_db(cr, m)
            upgrade_api = get_upgrade_api(m.db_id, m.source_version, pg_cnx, pg_env, token=m.request_key)
            req_status = upgrade_api.request_status(m.request_id, m.request_key).get('status', False)
            # if done, update
            if req_status == 'done':
                m = update_migration(m, state='upgrade_done')
            # if done, update
            elif req_status == 'warning':
                m = update_migration(m, state='upgrade_warning')
            # if prod migration and reasonable delay passed, cancel and unlock
            elif m.mode == 'self' and m.date_limit < datetime.datetime.utcnow():
                _migration_abort(m, 'time limit exceeded')
            # request archived/cancelled/failed on platform :'(
            elif req_status in ('archived', 'cancelled', 'failed'):
                _migration_abort(m, 'request %s on upgrade platform' % req_status)
            elif req_status and req_status.startswith('error:'):
                msg = "Error when querying status for migration #{0} (db: {1}): {2}"
                _logger.error(msg.format(m.id, m.db_id.name, req_status))

@log_call
def process_cleanup():
    with cursor() as cr:
        # Cleanup filestore archives
        request_fs_archives = {
            f: os.path.basename(f).split('-')[0]
            for f in glob.glob(os.path.join(TMPMIGDIR, '*.filestore.zip'))
        }
        if request_fs_archives:
            # Get migrations for which it's legit there is still a filestore archive
            cr.execute("""
                SELECT array_agg(request_id::varchar)
                FROM migrations
                WHERE request_id IN %s
                AND state_filestore != 'done'
                AND state NOT IN ('cancelled', 'failed')
                AND now() at time zone 'UTC' < create_date + interval '7 days'
            """, [tuple(request_fs_archives.values())])
            valid_request_ids = cr.fetchone()[0] or []
            # And then remove the non-legit filestore archives
            for archive_path, request_id in request_fs_archives.items():
                if request_id not in valid_request_ids:
                    _logger.info("Filestore cleanup for request %s : %s", request_id, archive_path)
                    os.unlink(archive_path)
                    symlink_path = os.path.join(TMPMIGSTATICDIR, os.path.basename(archive_path))
                    if os.path.islink(symlink_path):
                        _logger.info("Filestore cleanup (symlink) for request %s : %s", request_id, symlink_path)
                        os.unlink(symlink_path)


def zip_dir(path, stream, include_dir=True, fnct_sort=None, compression=zipfile.ZIP_DEFLATED):
    """
        Redefinition of zip_dir from odoo.tools.osutil to allow setting a custom compression
        level during the zip file generation.
    """
    path = os.path.normpath(path)
    len_prefix = len(os.path.dirname(path)) if include_dir else len(path)
    if len_prefix:
        len_prefix += 1

    with zipfile.ZipFile(stream, 'w', compression=compression, allowZip64=True) as zipf:
        for dirpath, dirnames, filenames in os.walk(path):
            filenames = sorted(filenames, key=fnct_sort)
            for fname in filenames:
                bname, ext = os.path.splitext(fname)
                ext = ext or bname
                if ext not in ['.pyc', '.pyo', '.swp', '.DS_Store']:
                    path = os.path.normpath(os.path.join(dirpath, fname))
                    if os.path.isfile(path):
                        zipf.write(path, path[len_prefix:])
        zip_result = zipf.testzip()
        if zip_result:
            raise zipfile.BadZipfile(zip_result)
        _logger.info("Zip %s test sucessful", stream)

def prepare_filestore():
    with cursor() as cr:
        # PreFlights doesn't need filestore
        cr.execute("""
            UPDATE migrations
               SET state_filestore=NULL
             WHERE state_filestore = 'pending'
               AND request_key IS NOT NULL
               AND mode IN ('rr-preflight','rr-preflight-muted')
        """)

        cr.execute("""
            SELECT *
              FROM migrations
             WHERE state_filestore = 'pending'
               AND request_key IS NOT NULL
        """)
        for mig in cr.fetchall():
            mig = _link_db(cr, mig)
            update_migration(mig, state_filestore='progress')
            try:
                token = generate_token(mig.request_id)
                archive_path = os.path.join(TMPMIGDIR, '%s-%s.filestore.zip' % (mig.request_id, token))
                fs = os.path.join(os.path.expanduser('~/filestore'), mig.db_id.name)
                zip_dir(fs, archive_path, compression=zipfile.ZIP_STORED)
                if not os.path.isdir(TMPMIGSTATICDIR):
                    os.mkdir(TMPMIGSTATICDIR)
                smlnk = os.path.join(TMPMIGSTATICDIR, os.path.basename(archive_path))
                if os.path.exists(smlnk):
                    os.unlink(smlnk)
                os.symlink(archive_path, smlnk)
                update_migration(mig, state_filestore='ready')
            except Exception:
                _logger.exception("migration: Bang! Cannot prepare filestore for database %s", mig.db_id.name)
                update_migration(mig, state_filestore='failed')
                _migration_abort(mig, "Cannot prepare filestore")

def remote_restore_init():
    """ Loop over migrations in upgrade_done and request a restore to master. """
    query = """
        SELECT *
          FROM migrations
         WHERE state = 'upgrade_done'
           AND state_filestore = 'ready'
           AND request_id is NOT NULL
    """
    restore_req_url = MASTER + '/saas_master/restore/request'
    verify = not TEST
    with cursor() as cr:
        cr.execute(query)
        for mig in cr.fetchall():
            mig = _link_db(cr, mig)
            (backup_size, fs_size) = _get_size(mig.db_id.name)
            params = {
                'dbname': mig.db_id.name,
                'mode': mig.mode,
                'source_version': mig.source_version,
                'target_version': mig.target_version,
                'request_id': mig.request_id,
                'request_key': mig.request_key,
                'purpose': 'upgrade',
                'email': mig.email,
                'dbuuid': mig.db_id.uuid,
                'extra_apps': mig.db_id.extra_apps,
                'size_db': int(backup_size) // (1024 * 1024 * 1024),
                'size_filestore': int(fs_size) // (1024 * 1024 * 1024),
            }
            # Increased the timeout to 60 following a bunch number of cases reaching the 15 secondes timeout
            res = requests.post(restore_req_url, data=params, verify=verify, auth=SaasAuth(), timeout=60)
            if res.status_code == 200:
                _logger.info('migration: %s will be restored on %s in version %s', mig.db_id.name, res.json()['server'], mig.target_version)
                update_migration(mig, state='done')
            else:
                _logger.error('migration: error when requesting a restore of %s', mig.db_id.name)


def local_restore_queue():
    """ Loop over migrations in upgrade_to_restore, download and restore """
    query = """
        SELECT *
          FROM migrations
         WHERE state = 'upgrade_to_restore'
           AND request_id is NOT NULL
    """
    with cursor() as cr:
        cr.execute(query)
        for mig in cr.fetchall():
            mig = _link_db(cr, mig)
            upgrade_api = get_upgrade_api(mig.db_id, mig.source_version, pg_cnx, pg_env, token=mig.request_key)
            if upgrade_api._dump_format == 'zip':
                fd, archfile = tempfile.mkstemp(dir=TMPMIGDIR)
                os.close(fd)
            else:
                archfile = tempfile.mkdtemp(dir=TMPMIGDIR)
            try:
                status = upgrade_api.request_status(mig.request_id, mig.request_key).get('status', False)
                if status == "archived":
                    _migration_abort(mig, 'The upgrade request has been archived')
                    update(mig.db_id, status='drop', expire_date="now() at time zone 'UTC'")
                    continue
                upgrade_api.download_database(mig.request_id, mig.request_key, archfile)
                _restore(mig.db_id.name, archfile)
                prod = mig.mode == 'self'
                try:
                    token = generate_token(mig.request_id)
                    _fetch_remote_filestore(mig.db_id.name, mig.source_name, mig.request_id, token)
                except Exception:
                    _logger.error(
                        "migration: Cannot fetch filestore for db %s. Retry with token of previous timeslot",
                        mig.db_id.name,
                    )
                    try:
                        token = generate_token(mig.request_id, timestep=-1)
                        _fetch_remote_filestore(mig.db_id.name, mig.source_name, mig.request_id, token)
                    except Exception:
                        if prod:
                            raise
                        _logger.exception("migration (copy): unable to fetch filestore for database %s", mig.db_id.name)
                _post_migrate(mig.db_id.name, mig.target_version, prod=prod)
            except Exception:
                _logger.exception("migration: Bang! Cannot restore database %s", mig.db_id.name)
                _cleanup_migration(mig.db_id.name)
                continue

            finally:
                if os.path.isdir(archfile):
                    shutil.rmtree(archfile)
                else:
                    os.unlink(archfile)
            expire_date = max(
                mig.db_id.expire_date,
                datetime.datetime.now() + datetime.timedelta(days=1),
            )
            update(mig.db_id, status='activated', expire_date=expire_date.isoformat())
            reset_cron_round(mig.db_id.name, resettime=True)

            if mig.mode == 'self':
                mig = update_migration(mig, state='finalize_prod')
            else:
                mig = update_migration(mig, state='done')

            if mig.mode != 'internal':
                migration_notify(mig)

def _migration_abort(mig, error):
    m = update_migration(mig, state='failed')
    msg = 'Error when processing migration #{0} (db: {1}): {2}.'
    if m.mode == 'self':
        update(m.db_id, status=None)
        reset_cron_round(m.db_id.name)
        msg += ' Access to db has been restored'
    _logger.error(msg.format(m.id, m.db_id.name, error))
    migration_notify(m)

def finalize_prod_migration():
    """
        Loop over migrations in finalize_prod,
        change dns in master
        and notify source trial that migration is successful
    """
    query = """
        SELECT *
          FROM migrations
         WHERE state = 'finalize_prod'
    """
    with cursor() as cr:
        cr.execute(query)
        for mig in cr.fetchall():
            mig = _link_db(cr, mig)
            # In P2, res.content is a str while in P3 it is a bytes. Note that the check is done in this
            # order on purpose. Indeed:
            # - P2: isinstance(b'test', str) = True and isinstance('test', bytes) = True
            # - P3: isinstance(b'test', str) = False and isinstance('test', bytes) = False
            # By testing isinstance(b'test', str) first, the test should pass only in P2, but not in P3.
            try:
                res = requests.get(
                    '{0}/saas_master/hostof?dbname={1}'.format(MASTER, mig.db_id.name), auth=SaasAuth(), timeout=MASTER_TIMEOUT
                )
                res.raise_for_status()
            except Exception:
                _logger.exception('Migration #%s (db: %s): error when fetching the source host', mig.id, mig.db_id.name)
                continue

            if isinstance(res.content, str):
                source_host = res.content
            else:
                source_host = res.content.decode()

            book_url = '{}/saas_master/dn/saas/ensure'.format(MASTER)
            with open(os.path.expanduser('~/config/config.json')) as fp:
                config = json.load(fp)
                host = config['me']['name']
            params = {
                'name': '{}.{}'.format(mig.db_id.name, SAAS_BASE_URL),
                'host': host,
            }
            verify = not TEST
            _logger.info('production migration: change DNS in master to point db %s to this server', mig.db_id.name)
            retry = 1
            while retry <= 5:
                try:
                    book_res = requests.post(
                        book_url, json={'params': params}, verify=verify,
                        auth=SaasAuth(), timeout=MASTER_TIMEOUT
                    )
                    book_res.raise_for_status()
                    break
                except Exception:
                    _logger.exception(
                        "Migration #%s (db: %s): error when booking the domain name. Attempt %s/5",
                        mig.id,
                        mig.db_id.name,
                        retry,
                    )
                wait_time = random.uniform(0.0, 2 ** retry)
                retry += 1
                time.sleep(wait_time)
            else:
                # cannot change DNS, process the next migration
                continue

            mig = update_migration(mig, state='done')

            # notify source server that migration is done
            succes_url = 'https://{}/saas_worker/migration-success'.format(source_host)
            params = {
                'dbname': mig.db_id.name,
                'dbuuid': mig.db_id.uuid,
                'source_version': mig.source_version,
                'request_key': mig.request_key,
            }
            # if old db failed to rename, it remains unreachable since dns is correct
            # in master, at worst the old db is unacessible...
            # Do not fail if migration-success fails or if any of the following steps fails
            retry = 1
            while retry <= 5:
                try:
                    _logger.info(
                        "production migration: notify source database %s on server %s that it can rename itself for backup",
                        mig.db_id.name,
                        source_host,
                    )
                    res = '?'
                    res = requests.post(succes_url, json={'params': params}, verify=verify, auth=SaasAuth(), timeout=MASTER_TIMEOUT)
                    res.raise_for_status()
                    succes_json = res.json()
                    result = succes_json.get('result', {})
                    db_limits = result.pop('db_limits', {})
                    db_stats = result.pop('db_stats', {})
                    db_domains = result.pop('db_domains', {})
                    _logger.info('production migration: source server response for original db rename: %s', succes_json)

                    # Copy the email and print limits from the source database
                    limits = {k: v for k, v in db_limits.items() if k in DB_LIMIT_FIELDS}
                    if 'expire_date' in limits.keys():
                        if (db_limits.get('expire_date') or '') < (datetime.datetime.now() + datetime.timedelta(days=30)).isoformat():
                            limits['expire_date'] = "now() at time zone 'UTC' + interval '30 days'"
                    update(mig.db_id, **limits)
                    # Duplicate the db_stats from the source database to keep history
                    _update_table_from_json(mig.db_id.id, 'db_stats', DB_STAT_FIELDS, db_stats)
                    _update_table_from_json(mig.db_id.id, 'domains', DB_DOMAIN_FIELDS, db_domains)
                    break
                except Exception:
                    _logger.exception(
                        "production migration: error when calling migration-success for db %s. Attempt %s/5\nRepsonse: %s",
                        mig.db_id.name,
                        retry,
                        res,
                    )
                wait_time = random.uniform(0.0, 2 ** retry)
                retry += 1
                time.sleep(wait_time)
            else:
                continue


def _createdb(dbname):
    """
    Create a database.

    :param str dbname: database name
    :raises Exception: if the operation failed
    """
    rc = pg_call('createdb --template="template0" "{}"'.format(dbname))
    if rc:
        raise Exception("Could not create database %s" % dbname)

def _perm_filestore(fs):
    """
    Ensure all files and directories are readable by 'other' (necessary for nginx)
    """
    for root, _, files in os.walk(fs):
        _logger.debug('set permissions of dir %s', root)
        os.chmod(root, 0o755)
        for f in files:
            _logger.debug('set permissions of file %s', os.path.join(root, f))
            os.chmod(os.path.join(root, f), 0o644)

def _sync_filestore(fs, dbname):
    """
    Sync filestore

    :param str fs: filestore path
    :param str dbname: database name
    """
    destdir = os.path.join(os.path.expanduser('~/filestore'), dbname)
    if not os.path.isdir(destdir):
        os.mkdir(destdir)
    if os.path.isdir(fs):
        fs_cmd = "rsync -a {} {}".format(os.path.join(fs, ''), os.path.join(destdir, ''))
        subprocess.call(fs_cmd, shell=True)

def _restore(dbname, archfile):
    """ create database and restore using archfile.
        archfile is supposed to be a zip file.
    """
    _createdb(dbname)
    if os.path.isdir(archfile):
        # directory dump
        # restore db
        dump = os.path.join(archfile, "upgraded.dump")
        subprocess.call(["pg_restore", "-O", "-F", "d", dump, "-d", dbname, "-j", "4"])
        # copy filestore if provided
        fs = os.path.join(archfile, 'filestore')
        _perm_filestore(fs)
        _sync_filestore(fs, dbname)
    else:
        # zip, legacy request
        destdir = tempfile.mkdtemp(dir=TMPMIGDIR)
        with zipfile.ZipFile(archfile, 'r') as z:
            try:
                z.extractall(destdir)
                # restore db
                psql_cmd = "psql -q -d {} -f {}".format(dbname, os.path.join(destdir, 'dump.sql'))
                subprocess.call(psql_cmd, shell=True)
                # copy filestore if provided
                fs = os.path.join(destdir, 'filestore')
                _perm_filestore(fs)
                _sync_filestore(fs, dbname)
            finally:
                shutil.rmtree(destdir)


def _fetch_remote_filestore(local_name, remote_name, request_id, token):
    """ fetch remote filestore and sync it with the local one.
        (because the source filestore isn't uploaded to the upgrade platform
        to begin with)
    """
    if not remote_name:
        remote_name = local_name
    static_route = os.path.join(*TMPMIGSTATICDIR.partition('saas_worker/')[1:])
    source_host = requests.get(
        '{0}/saas_master/hostof?dbname={1}'.format(MASTER, remote_name),
        auth=SaasAuth(), timeout=TIMEOUT
    ).content.decode('UTF-8')
    url = '{}://{}/{}/{}-{}.filestore.zip'.format(SCHEME, source_host, static_route, request_id, token)
    verify = not TEST
    fd, archfile = tempfile.mkstemp(dir=TMPMIGDIR)
    os.close(fd)
    destdir = tempfile.mkdtemp(dir=TMPMIGDIR)
    try:
        res = requests.get(url, stream=True, verify=verify, auth=SaasAuth(), timeout=7200)
        res.raise_for_status()
        with open(archfile, 'wb') as fh:
            for chunk in res.iter_content(chunk_size=1024):
                fh.write(chunk)

        fs = os.path.join(os.path.expanduser('~/filestore'), local_name)
        if not os.path.isdir(fs):
            os.mkdir(fs)

        with zipfile.ZipFile(archfile, 'r', allowZip64=True) as z:
            z.extractall(path=destdir)
        # Delete archive early to reduce the risk of out-of-space errors
        os.unlink(archfile)
        rsync_cmd = ["rsync", "-a", os.path.join(destdir, remote_name, ''), os.path.join(fs, '')]
        subprocess.check_call(rsync_cmd)
        url = '{}://{}.{}/saas_worker/s2s/filestore'.format(SCHEME, remote_name, SAAS_BASE_URL)
        requests.get(url, verify=verify, json={'params': {'request_id': request_id}}, auth=SaasAuth(), timeout=TIMEOUT)
    finally:
        if os.path.exists(archfile):
            os.unlink(archfile)
        shutil.rmtree(destdir)


def _cleanup_migration(dbname):
    drop_conn(dbname)
    pg_call('dropdb --if-exists "{0}"'.format(dbname))
    fs = os.path.join(os.path.expanduser('~/filestore'), dbname)
    if os.path.isdir(fs):
        shutil.rmtree(fs)


def _post_migrate(dbname, to, prod=False):
    """ post migration/restore queries. """
    db = get(dbname)
    # not a one-app free if testing
    if not prod:
        update(db, extra_apps=True)
    uuid = db.uuid
    hostname = '%s.%s' % (dbname, SAAS_BASE_URL)

    queries = [
        "CREATE EXTENSION IF NOT EXISTS unaccent",
        "UPDATE ir_config_parameter SET value='%s' WHERE key='database.uuid'" % uuid,
        "UPDATE ir_config_parameter SET value='%s://%s' WHERE key='web.base.url'" % (SCHEME, hostname),
    ]
    if prod:
        # keep custom domain: only update if the domain ends in .odoo.com
        queries.append("UPDATE ir_config_parameter SET value='%s' WHERE key='mail.catchall.domain' AND value like '%%.odoo.com'" % hostname.split(':')[0])
    else:
        queries.append("UPDATE ir_config_parameter SET value='%s' WHERE key='mail.catchall.domain'" % hostname.split(':')[0])
        queries.append("UPDATE ir_cron SET active=false")

    queries.append(
        """UPDATE ir_cron SET active=true
            WHERE id IN (
                SELECT res_id FROM ir_model_data
                WHERE model='ir.cron' AND (module, name) IN (
                    ('mail', 'ir_cron_module_update_notification'),
                    ('base', 'autovacuum_job')
                )
            )"""
    )

    has_oauth = bool(pg_call("""\
        psql -tAqc "SELECT count(1) FROM information_schema.columns WHERE table_name='auth_oauth_provider' AND column_name='client_id'" {0} | xargs -- bash -c 'exit $0'
    """.format(dbname)))
    if has_oauth:
        queries.append(
            "UPDATE auth_oauth_provider SET client_id='%s' WHERE id=(SELECT res_id FROM ir_model_data WHERE model='auth.oauth.provider' AND name='provider_openerp')" % uuid
        )

    pg_call('psql -q -d {0} -c "{1};"'.format(dbname, ';'.join(queries)))

    _force_ping(dbname, to)


def _force_ping(dbname, to):
    db = get(dbname)
    if db and db.uuid in _NO_PING:
        _logger.warning("Skip ping for database %r", dbname)
        return
    ping = "env['publisher_warranty.contract'].browse([]).update_notification(cron_mode=False);env.cr.commit()"
    lh = "--log-handler=:ERROR --log-handler=odoo.addons.mail:DEBUG"
    subprocess.call(
        'echo "{2}" | ~/bin/openerp-{0} shell -d {1} {3} --logfile={4}'.format(to, dbname, ping, lh, _log_file),
        shell=True)

def _update_module_list(db):
    assert isnamedtuple(db)
    version = get_branch(db.port)
    update = "env['ir.module.module'].browse([]).update_list();env.cr.commit()"
    subprocess.call(
        'echo "{2}" | timeout --signal=KILL 30s ~/bin/openerp-{0} shell -d {1} --log-handler=:ERROR'.format(version, db.name, update),
        shell=True)


def stats():
    # TODO callable from oe-meta cli
    # TODO percentage per port
    s = {
        'trial': 0,
        'paid': 0,
        'drop': 0,
        'count': 0,
        'all_domains': [],
        'custom_domains': [],
    }
    with cursor() as cr:
        modes = ('trial', 'paid', 'drop')
        cr.execute("""SELECT mode, COUNT(id)
                        FROM databases
                       WHERE mode in %s
                         AND COALESCE(status, '') != 'frozen'
                    GROUP BY mode
                   """, [modes])
        s.update(dict(cr.fetchall()))
        s['count'] = sum(s[x] for x in 'trial paid drop'.split())

        cr.execute("SELECT name FROM domains")
        s['custom_domains'] = [d[0] for d in cr.fetchall()]

        cr.execute("SELECT name FROM databases WHERE mode IN %s", [modes])
        doms = list(map(''.join, itertools.product((x[0] for x in cr.fetchall()), SAAS_VALID_SUFFIXES)))
        s['all_domains'] = doms + s['custom_domains']

    return s


def print_db(db, print_hosts=True, as_json=False):
    if not db:
        return
    assert isnamedtuple(db)

    if as_json:
        d = db._asdict()
        d.pop('id', None)
        if print_hosts:
            d['hosts'] = get_hosts(db)
        dthandler = lambda obj: obj.isoformat() if isinstance(obj, datetime.datetime) else None
        print(json.dumps(d, default=dthandler))
    else:
        if print_hosts:
            print(db, 'hosts:', get_hosts(db))
        else:
            print(db)


def reset_cron_round(dbname, resettime=False):
    with cursor() as cr:
        cr.execute("SELECT cron_round FROM databases WHERE name=%s", [dbname])
        [old] = cr.fetchone() or [None]
        if old is None:
            return None, None
        crontime = 'NULL' if resettime else 'cron_time'
        cr.execute("""UPDATE databases d
                         SET cron_round = (SELECT COALESCE(MIN(cron_round), 0)
                                             FROM databases
                                            WHERE mode IN %s
                                              AND coalesce(status, '') NOT IN %s
                                              AND port = d.port
                                              AND id != d.id)
                           , cron_time={0}
                       WHERE name = %s
                   RETURNING cron_round
                   """.format(crontime), [ACTIVE_MODES, DISABLED_STATUSES, dbname])
        [new] = cr.fetchone() or [None]
    return old, new


def dumps(dbname):
    db = get(dbname)
    if not db:
        return None
    d = db._asdict()
    d.pop('id', None)
    if db.source_uuid:
        d['source_uuid'] = db.source_uuid.uuid
    d['hosts'] = get_hosts(db)
    dthandler = lambda obj: obj.isoformat() if isinstance(obj, datetime.datetime) else None
    j = json.dumps(d, default=dthandler, separators=(',', ':'), sort_keys=1)
    return base64.urlsafe_b64encode(zlib.compress(j.encode())).decode().strip('=')


def loads(data):
    if not data:
        return
    zdata = base64.urlsafe_b64decode(data.encode() + b'=' * (-len(data) % 4))
    j = json.loads(zlib.decompress(zdata).decode())
    dbname = j.pop('name')
    hosts = j.pop('hosts', [])
    db = get(dbname)
    if db:
        update(db, **j)
    else:
        add(name=dbname, **j)
        db = get(dbname)

    if hosts:
        with cursor() as cr:
            for host in hosts:
                add_host(cr, db, host)

    reset_cron_round(dbname)


@log_call
def rr_preflight_process():
    """
    For the upgrades that have been completed and were started as a pre-flight request, fetch
    the migration metadata (time + disabled views) and update the migration record so user can
    (potentially) receive a popup.
    """
    with cursor() as cr:
        cr.execute("""
            SELECT *
              FROM migrations
             WHERE state IN ('upgrade_done', 'upgrade_warning')
               AND request_key IS NOT NULL
               AND mode IN ('rr-preflight', 'rr-preflight-muted')
        """)
        for mig in cr.fetchall():
            mig = _link_db(cr, mig)
            upgrade_api = get_upgrade_api(mig.db_id, mig.source_version, pg_cnx, pg_env, token=mig.request_key)
            req_datas = upgrade_api.request_status(mig.request_id, mig.request_key)
            update_migration(mig, state='rolling_release_available' if mig.state == 'upgrade_done' else 'rolling_release_warning', state_filestore='', rr_estimated_time=req_datas.get('elapsed_time'), disabled_view_count=req_datas.get('disabled_view_count'))
            # migration_notify(mig)


def get_rr_status(dbname):
    """
    If there was a preflight (not rr-preflight-muted) done for the DB, retrieve the associated collected info
    """
    with cursor() as cr:
        cr.execute(r"""
     WITH upgrades AS (
       SELECT *, COALESCE('{"7.0": 0, "8.0": 5.5, "9.0": 6.5, "10.0": 13.5, "11.0": 18.5}'::jsonb->>source_version,
                          regexp_replace(source_version, '\D', '', 'g'))::float8 as branch_key
         FROM migrations
        WHERE mode='rr-preflight'
     )
     SELECT u.*
       FROM upgrades u
       JOIN databases d ON u.db_id = d.id
                       AND d.port = CASE WHEN u.branch_key < 100 THEN (8069+100*u.branch_key)::int
                                         ELSE 42000 + u.branch_key::int
                                     END
      WHERE d.name=%s
        AND (d.rr_locked_until IS NULL OR d.rr_locked_until < (CURRENT_DATE + INTERVAL '30 days'))
   ORDER BY u.id DESC
      FETCH FIRST ROW ONLY
        """, [dbname, ])
        m = cr.fetchone()
        if m:
            return {
                'time': m.rr_estimated_time,
                'state': m.state,
                'target_version': m.target_version,
                'source_version': m.source_version,
                'disabled_view_count': m.disabled_view_count,
                'leave_lts': m.source_version[0:4] != 'saas' and m.target_version[0:4] == 'saas',
                'remain_lts': m.source_version[0:4] != 'saas' and m.target_version[0:4] != 'saas',
                'is-edu': dbname.startswith(EDU_PREFIX),
            }
    return {}


def set_rr_dismissed(dbname):
    """
    If there was a preflight (not rr-preflight-muted) done for the DB, set it as dismissed
    and stop pinging user with it
    """
    if get_rr_status(dbname).get('leave_lts'):
        db = get(dbname)
        with cursor() as cr:
            cr.execute("update migrations SET state='rr-dismiss' WHERE db_id=%s and mode='rr-preflight'", (db.id,))
        return {'requested': True}
    return {'requested': False}

def rr_remove_from_queue(dbname):
    """
    Remove the DB from the RR queue
    if state= queued and mode = rr-forced
    """
    with cursor() as cr:
        cr.execute("""
            UPDATE migrations
               SET state='cancelled'
             WHERE db_id = (SELECT id FROM databases WHERE name=%s)
               AND state='queued'
               AND mode='rr-forced'
        """, [dbname, ])



def rr_mass_test_migrate(from_version, to_version, muted=False, include_failed=False, clean=False, all=False):
    """
    Create migration request (in mode of "rr-preflight" or "rr-preflight-muted") for all existing databases
    that don't have one yet for the given 'from' version.
    When "muted", the web client will not show the upgrade popup.
    When "clean", previous preflights for same from_version are removed.
    When "all", even db with a previous test migration are processed. Prepare the work for the forced migrations
    """
    with cursor() as cr:
        if muted:
            mode = 'rr-preflight-muted'
            inselect_condition = ""
        else:
            mode = 'rr-preflight'
            inselect_condition = " AND mode!='rr-preflight-muted'"
        from_port = get_port(from_version)
        if clean:
            rr_mass_test_clean(from_version)
        if all:
            inselect_condition = " AND FALSE"
        elif include_failed:
            inselect_condition += " AND state IN ('rolling_release_available', 'rolling_release_warning') AND NOT EXISTS (SELECT 1 FROM migrations m2 WHERE m2.db_id=migrations.db_id AND m2.mode='rr-preflight' AND m2.id>migrations.id)"
        for db in _each(
            cr,
            id=sql.not_inselect(cr.mogrify("select db_id from migrations where source_version=%s " + inselect_condition, [from_version]).decode()),
            mode=sql.in_('paid', 'trial'),
            port=from_port,
            status=sql.in_('open', 'pending', 'activated', '', None)
        ):
            if db.name.startswith(EDU_PREFIX):
                continue  # No edu databases
            add_migration(db, actuator="meta", mode=mode, source_version=from_version, target_version=to_version)


def rr_mass_test_clean(source_version=None, target_version=None):
    """
    Clean all preflights migrations.
    @param source_version: source version to clean.
    @param target_version: target version to clean.
    If no version is given (--all), all preflights are deleted.
    """
    with cursor() as cr:
        query = """
            DELETE FROM migrations
             WHERE mode in ('rr-preflight', 'rr-preflight-muted')
               AND state!='rr-dismiss'
        """
        params = []
        if source_version:
            query += " AND source_version=%s"
            params.append(source_version)
        if target_version:
            query += " AND target_version=%s"
            params.append(target_version)
        cr.execute(query, params)


def rr_forced_process_queue():
    """
    Process rr-forced one at a time
    Simply switch mode from rr-forced to self as if it was a regular production upgrade
    Ensure no other production upgrade is in progress before switching one
    This makes the queue handled by the saas server and not the upgrade server,
    to lower as much as possible the locking time of the database
    """
    if not geoip2:
        sys.exit("geoip2 is not available on this system")
    config = json.load(open("/home/odoo/config/config.json", "rb"))
    gip = geoip2.database.Reader("/usr/share/GeoIP/GeoLite2-City.mmdb")
    server_tz = tz.gettz(gip.city(config["me"]["public_ip"]).location.time_zone)
    if datetime.datetime.now(server_tz).hour < 7 or datetime.datetime.now(server_tz).hour >= 21:
        with cursor() as cr:
            cr.execute("""
                WITH db_to_upgrade AS (
                    SELECT m.id
                      FROM migrations m
                      JOIN databases d ON m.db_id = d.id
                     WHERE m.mode='rr-forced'
                       AND (d.rr_locked_until IS NULL OR d.rr_locked_until < CURRENT_DATE)
                       AND NOT EXISTS (SELECT 1
                                         FROM migrations
                                        WHERE mode='self'
                                          AND state='queued'
                                      )
                     FETCH FIRST ROW ONLY
                )
                UPDATE migrations m
                   SET mode='self',
                       state_filestore='pending',
                       create_date=now() at time zone 'UTC'
                  FROM db_to_upgrade d
                 WHERE m.id=d.id
            """)
    return True


def rr_preflight_unmute(from_version, to_version):
    """
    Muted preflights are easy for testing migrations prior release of the given version on the SaaS
    When version is released on the SaaS, unmute them make all the tested upgrades instantly available
    """
    with cursor() as cr:
        cr.execute("""
            UPDATE migrations
               SET mode='rr-preflight'
             WHERE mode='rr-preflight-muted'
               AND source_version=%s
               AND target_version=%s
        """, (from_version, to_version))

def unlock_maintenance_mig():
    """
    mode='maintenance_mig' AND state='done' is reached when the remote restore is scheduled
    If, for any reason, request is archived on the upgrade platform, it will not be restored on target.
    We wait for a 10 minutes delay before unlocking the request, to prevent race condition
    (remote restore start->archiving->unlock->remote restore end-> split brain! )
    """
    with cursor() as cr:
        ten_minutes = (datetime.datetime.now() - datetime.timedelta(minutes=10)).isoformat()
        cr.execute("""
            SELECT m.*
              FROM migrations m
              JOIN databases d ON m.db_id=d.id
             WHERE d.status='maintenance_mig'
               AND m.state='done'
               AND NOT EXISTS (SELECT 1 from migrations m2 where m.db_id=m2.db_id and m.id<m2.id)
        """)
        for upg in cr.fetchall():
            m = _link_db(cr, upg)
            upgrade_api = get_upgrade_api(m.db_id, m.source_version, pg_cnx, pg_env, token=m.request_key)
            req_status = upgrade_api.request_status(m.request_id, m.request_key)
            if upg.date_limit:
                dl = min(req_status['write_date'], upg.date_limit.isoformat())
            else:
                # No date limit should not be possible.
                # As preflight are created some days prior their effective execution
                # we should add some days to compute a limit.
                dl = min(req_status['write_date'], (upg.create_date + datetime.timedelta(days=10)).isoformat())
            if req_status['status'] == 'archived' and dl < ten_minutes:
                reset_cron_round(m.db_id.name)
                update(m.db_id, status=None)


def main():
    """
        Usage:
            oe-meta set [-C] [-p <port> | -b <branch>] [-m <mode>] [-s <status>] [-H <host>...] [-c <create_date>] [-e <expire_date>] [-l <email_daily_limit>] [-a | -A] [-u <uuid>] <database>
            oe-meta lock [-M | -g] <database>
            oe-meta block <database>
            oe-meta abuse <database>
            oe-meta (unlock | unblock) <database>
            oe-meta get [-j] <database>
            oe-meta (list | count) [-f] [-m <mode>...] [-s <status>...] [-o <order>] [<version>]
            oe-meta list-branches [-a] [-p]
            oe-meta last-branch [-a]
            oe-meta last-port [-a]
            oe-meta get-branch <port>
            oe-meta get-port <branch>
            oe-meta migrate [-v...] [-f] [--keep-failed] [--copy] [--inplace] [--no-update-list] [--upgrade-path <upgrade_path>] <database> [<version>]
            oe-meta auto-migrate <from-version> <to-version>
            oe-meta auto-migrate-db <database> <to-version> [<mode>]
            oe-meta queue-migrate [-l <limit>]
            oe-meta reset-cron-round [-t] <database>
            oe-meta dumps <database>
            oe-meta loads <data>
            oe-meta mass-test-migrate [--mute] [--include-failed] [--clean] [--all] <from-version> <to-version>
            oe-meta mass-test-clean [<to-version>] [<from-version>]
            oe-meta preflight-unmute <from-version> <to-version>

        Options:
            -m <mode>
            -s <status>
            -l <limit>
            -o <order>                  [default: name]
            -x <exclude>
            -X <exclude_file>
            -H <host>
            -c <create_date>
            -e <expire_date>
            -u <uuid>
            -a                          extra_apps (default)
            -A                          no extra_apps, i.e. freemium

            -C                          Create entry if database not found
            --start-at <start-at>
            --stop-at <stop-at>

            -n, --dry-run
            -k, --keep-failed

            --copy                     keep original DB intact and migrate a copy of the DB
            --inplace                  migrate database without making a temp copy
                                       if migration fails, original database is unusable
                                       cannot be used whith --copy

            --mute                     preflight migration, without showing it to user even if success
            --include-failed           preflight migration, include previous failed preflight in batch
            --clean                    preflight migration, remove previous pre-flight results
            --all                      preflight migration, process all db, even with previous tests,
                                       to prepare the auto-migrate work (forced migrations)
    """
    from docopt import docopt
    import signal

    if os.getuid() == 0:
        sys.exit("oe-meta cannot be run as root")

    signal.signal(signal.SIGINT, _sigint_handler)

    opt = docopt(main.__doc__)

    write_cmds = 'set lock block abuse unlock unblock migrate auto-migrate queue-migrate reset-cron-round loads'.split()
    if sys.argv[1] in write_cmds:
        _logger.info('call oe-meta %s', subprocess.list2cmdline(sys.argv[1:]))

    if any(opt[m] for m in 'migrate auto-migrate queue-migrate'.split()):
        hndl = logging.StreamHandler()
        hndl.setFormatter(_log_fmt)
        _logger.addHandler(hndl)

    if opt['set']:
        dbname = opt['<database>']
        create = opt['-C']
        kw = {}

        for o in 'create_date expire_date uuid'.split():
            v = opt['-' + o[0]]
            if v:
                kw[o] = v
        for o in 'mode status'.split():
            v = opt['-' + o[0]]
            if v:
                kw[o] = v[0]
        for o in 'port'.split():
            v = opt['<%s>' % o]
            if v:
                kw[o] = v

        if opt['-l']:
            kw['email_daily_limit'] = opt['-l']
        if opt['-a']:
            kw['extra_apps'] = True
        if opt['-A']:
            kw['extra_apps'] = False
        if opt['<branch>']:
            kw['port'] = get_port(opt['<branch>'])

        db = get(dbname)
        if kw:
            if db:
                update(db, **kw)
            else:
                if not create:
                    sys.exit('Unknown database: %s' % (dbname,))
                    return
                db = add(name=dbname, **kw)

        if opt['-H']:
            with cursor() as cr:
                for host in opt['-H']:
                    add_host(cr, db, host)

        reset_cron_round(dbname)

    elif any(opt[x] for x in 'lock block abuse unlock unblock'.split()):
        dbname = opt['<database>']
        reset_cron_round(dbname)
        status = None
        if opt['block']:
            status = 'maintenance_block'
        elif opt['lock']:
            status = 'maintenance_mig' if opt['-M'] else \
                     'maintenance' if opt['-g'] else \
                     'maintenance_block'
        elif opt['abuse']:
            status = 'abuse'
        with cursor() as cr:
            if status == 'abuse':
                # expire in 10 days to keep the weekly backup
                cr.execute(
                    """
                    UPDATE databases
                    SET mode='drop',
                        status='abuse',
                        expire_date=now() at time zone 'UTC' + interval '10 days'
                    WHERE name=%s
                    """, (dbname,))
            else:
                cr.execute("UPDATE databases SET status=%s WHERE name=%s", (status, dbname))
    elif opt['get']:
        d = get(opt['<database>'], mode=sql.neq(None))
        print_db(d, as_json=opt['-j'])
        sys.exit(not d)
    elif opt['list'] or opt['count']:
        kw = {'_order': opt['-o']}
        version = opt['<version>']
        if version is not None:
            kw['port'] = get_port(version)

        for k in 'mode status'.split():
            v = opt['-' + k[0]]
            if not v:
                continue
            kw[k] = sql.in_(*v)
        if 'mode' not in kw:
            kw['mode'] = sql.not_in('spare', 'hostname')

        if opt['count']:
            print(len(list(list_(**kw))))
            return

        full = opt['-f']
        for d in list_(**kw):
            if full:
                print_db(d)
            else:
                print(d.name)

    elif opt['list-branches']:
        for branch in list_branches(opt['-a']):
            if opt['-p']:
                print('%s (%d)' % (branch, get_port(branch)))
            else:
                print(branch)
    elif opt['last-port']:
        print(get_port(get_last_branch(opt['-a'])))
    elif opt['get-branch']:
        b = get_branch(int(opt['<port>']))
        if b:
            print(b)
    elif opt['get-port']:
        print(get_port(opt['<branch>']))
    elif opt['last-branch']:
        print(get_last_branch(opt['-a']))
    elif opt['migrate']:
        version = opt['<version>']
        if version is None:
            version = get_last_branch()
        try:
            migrate(opt['<database>'], version, opt['-v'], opt['-f'],
                    keep_failed=opt['--keep-failed'], copy=opt['--copy'],
                    inplace=opt['--inplace'],
                    update_list=not opt["--no-update-list"],
                    upgrade_path=opt["<upgrade_path>"] if opt["--upgrade-path"] else False,
            )
        except MigrationError as e:
            _logger.exception('MigrationError')
            sys.exit(e)
    elif opt['auto-migrate']:
        auto_migrate(opt['<from-version>'], opt['<to-version>'])
    elif opt['auto-migrate-db']:
        auto_migrate_db(opt['<database>'], opt['<to-version>'], opt.get('<mode>', False))
    elif opt['queue-migrate']:
        queue_migrate(migrations_to_handle=int(opt['-l'] or 10))
    elif opt['reset-cron-round']:
        db = opt['<database>']
        old, new = reset_cron_round(db, opt['-t'])
        print("%s: %s -> %s" % (db, old, new))
    elif opt['dumps']:
        d = dumps(opt['<database>'])
        if d:
            print(d)
    elif opt['loads']:
        data = opt['<data>']
        if data == '-':
            data = sys.stdin.read()
        loads(data)
    elif opt['mass-test-migrate']:
        rr_mass_test_migrate(opt['<from-version>'], opt['<to-version>'],
                             muted=opt.get('--mute', False),
                             include_failed=opt.get('--include-failed', False),
                             clean=opt.get('--clean', False),
                             all=opt.get('--all', False))
    elif opt['mass-test-clean']:
        rr_mass_test_clean(opt.get('<from-version>', None), opt.get('<to-version>', None))
    elif opt['preflight-unmute']:
        rr_preflight_unmute(opt['<from-version>'], opt['<to-version>'])



if __name__ == '__main__':
    main()
